"""
Created on Tue Jun  11 18:37:12 2020

@author: theok
"""

import numpy as np
import re
import matplotlib.pyplot as plt
from scipy.stats import ks_2samp, cumfreq, norm
import scipy
from scipy.spatial.distance import cdist
import copy
from scipy.stats import norm

def gaussian(mean, sigma, x1, x2):
    x=np.linspace(x1, x2, 1000)
    y=scipy.stats.norm(mean, sigma)
    pdf=y.pdf(x)
    cums=np.cumsum(pdf)/10.
    return (x, cums, pdf),(0., 0., 0.)


def calc_rmsd(data1, data2):
    diff=data1-np.array(data2)
    n=len(data1)
    Rmsd=np.sqrt((np.sum(diff**2)/n))
    return Rmsd

def calc_euclid(datax1, datay1, datax2,  datay2):
    mindists=[]
    for i in range(len(datax1)):
        x1, y1=datax1[i], datay1[i]
        point=[x1, y1]
        xdis, ydist=(x1-datax2)**2,  (y1-datay2)**2
        dist=np.sqrt(xdis + ydist)
        mindist=min(dist)
        mindists.append(mindist)
    return sum(mindists)
        

def get_coords(pdb, atomname='O1', chain='', resid='', resn='R1A'):
    coord_dict={}
    with open(pdb,'r') as f:
        lines=f.readlines()
        for line in lines:
            atom=re.search("^ATOM\s*\d*\s*%s\s*(%s)\s%s\s*%s(.*)" %(atomname, resn, chain, resid), line)
         
            if atom == None:
                continue
            else:
                #print atom.group(0)
                coords=[float(atom.group(2).split()[0]), float(atom.group(2).split()[1]), float(atom.group(2).split()[2])]
                coord_dict['resid %s and chain %s and resn %s and name %s' %(resid, chain, resn, atomname)]=coords

    return coord_dict
    
def calc_distance(dict_of_coords1, dict_of_coords2 ):
    for atom1, atom2 in zip(dict_of_coords1.keys(), dict_of_coords2.keys()):
        x1, x2= dict_of_coords1[atom1][0], dict_of_coords2[atom2][0]
        y1, y2= dict_of_coords1[atom1][1], dict_of_coords2[atom2][1]
        z1, z2= dict_of_coords1[atom1][2], dict_of_coords2[atom2][2]
        
        dist=np.sqrt(((x2-x1)**2) + ((y2-y1)**2) + ((z2-z1)**2))
    return dist
    
def calc_exp_hist(exp_hist, size):
    data=np.loadtxt(exp_hist)
    distances=data[:,0]
    prob=data[:,1]
    cumsum=np.cumsum(prob)/10.
    
    kde = scipy.stats.gaussian_kde(prob.T)
    p = 1 / kde.pdf(prob.T)
    p /= np.sum(p)
    
    idx = np.random.choice(np.arange(len(prob)), size=size, replace=False, p=p)
    idx=np.sort(idx)
    sample_prob = prob[idx]
    sample_dist=distances[idx]
    sample_cumsum=np.cumsum(sample_prob)/10.
    
    
    #print sample_prob
    #plt.plot(sample_dist, sample_cumsum, '-')
    #plt.show()
    return (distances, cumsum, prob), (sample_dist, sample_cumsum, sample_prob)
    
def read_pdbs(pdb_list):
    l=np.loadtxt(pdb_list, dtype=str)
    return l

def gen_ensemble(listOfPdbs, ensemble_size=5):
    ensemble=[]
    for i in range(int(ensemble_size)):
        str_index=np.random.randint(0, len(listOfPdbs))
        pdb=listOfPdbs[str_index]
        ensemble.append(pdb)
    return ensemble

def calc_ensemble_distances(ensemble, atomname='O1', resid='', resn='R1A'):
    distances=[]
    orig_ensemble=np.array(ensemble)
    for idx in range(len(ensemble)):
        structure=ensemble[idx]
        other_str_idx= np.delete(  np.arange(len(ensemble)), idx)
        for partner_id in other_str_idx:

            partner=ensemble[partner_id]
            #print partner, structure
            mtsl1=get_coords(structure, atomname=atomname, chain='A', resid=resid, resn=resn)
            mtsl2=get_coords(partner, atomname=atomname, chain='B', resid=resid, resn=resn)
            dist=calc_distance(mtsl1, mtsl2)
            distances.append(dist)

    return ensemble, distances

def fit(ensmble_dist, exp_dist, plot=False):
    #calc_dist=plt.hist(ensmble_dist[1], bins=len(ensemble), density=True, cumulative=True)[0]
    calc_dist, x1=np.histogram(ensmble_dist[1], bins=len(ensmble_dist[1]), density=True)
    calc_dist=np.cumsum(calc_dist)*(x1[1]-x1[0])
    #print ensmble_dist[1]

    
    real_distr=exp_dist[0]
    sampled_distr=exp_dist[1]

    d=calc_euclid(x1[1:] ,calc_dist, real_distr[0] ,real_distr[1] )
    
    if plot:
        ax=plt.subplot(121)
        
        ax.plot(x1[1:] ,calc_dist, '-')
        ax.plot(real_distr[0] ,real_distr[1] , '-')
        
        ax=plt.subplot(122)
        print len(ensmble_dist[1])/3
        plt.hist(ensmble_dist[1], bins=6, density=True)[0]
        ax.plot(real_distr[0], real_distr[2])


    return ensmble_dist[0], d


    
def evolve_ensemble(listOfPdbs, ensemble, exp_dist, mutations=1, min_members=5, max_members=10, rate=0.5,  mc_steps=100, stop_criterion=20 ):
    '''initial_ensemble is a calc_ensemble_distances object
    '''

    original_score=fit(ensemble, exp_dist)[1]
    original_ensemble=ensemble
    
    scores=[]
    for mc_step in range(mc_steps):
        
        if mc_step ==0:
            score_to_beat=original_score
            ensemble_to_beat=original_ensemble
            
        scores.append(score_to_beat)
      
        print   mc_step, score_to_beat, len(ensemble_to_beat[0]), len(ensemble_to_beat[1])
        
        old_ensemble = copy.deepcopy(ensemble_to_beat)
        
        for mutant in range(mutations):

            index_to_mutate=np.random.randint(0, len(old_ensemble[0]))
        
            new_structure_index= np.random.randint(0, len(listOfPdbs))
            new_structure=listOfPdbs[new_structure_index]
            #mutate
            old_ensemble[0][index_to_mutate]=new_structure
                        
            if np.random.uniform()<rate:
                if len(old_ensemble[0])< max_members:
                    add_structure_index= np.random.randint(0, len(listOfPdbs))
                    add_structure=listOfPdbs[add_structure_index]
                    old_ensemble[0].append(add_structure)
            
            if np.random.uniform()<rate:
                if len(old_ensemble[0])> min_members:
                    del_structure_index= np.random.randint(0, len(old_ensemble[0]))
                    del old_ensemble[0][del_structure_index]
           
            new_ensemble=calc_ensemble_distances(old_ensemble[0], atomname='O1', resid='82', resn='R1A')
            new_score=fit(new_ensemble, exp_dist)[1]
            
            
            if new_score<score_to_beat:
                score_to_beat=new_score
                ensemble_to_beat=new_ensemble
            else:
                continue

            #print 

            #fit(ensemble_to_bit, exp, plot=True)
            #plt.show()
            #plt.close()
            #try:
            #    try:
            #        print 'stop', score_to_bit, scores[mc_step-stop_criterion]
            #        np.testing.assert_almost_equal(score_to_bit, scores[mc_step-stop_criterion])
            #        print 'Stop criterion met'
            #        return ensemble_to_bit
            #    except AssertionError:
            #        continue
            #except IndexError:
            #    continue

    return ensemble_to_beat
        

members=7
exp=calc_exp_hist('DEER_1us_RPBdrug_1G_e_distribution.dat', size=members*(members-1))
pdbs=read_pdbs('pdb.list')

ensemble=gen_ensemble(pdbs, ensemble_size=members)
ensemble_dists=calc_ensemble_distances(ensemble, atomname='O1', resid='82', resn='R1A')
print ensemble_dists

opt_ensemble_dists=evolve_ensemble(pdbs, ensemble_dists, exp, min_members=4, max_members=10, mutations=1, mc_steps=50)

plt.figure(0)
print fit(ensemble_dists, exp, plot=True)

plt.figure(1)
#print opt_ensemble_dists
print 'Optimized ensemble:'
print fit(opt_ensemble_dists, exp, plot=True)
plt.show()



